package com.bureau.devicefingerprint

import android.Manifest
import android.content.Context
import android.content.pm.Signature
import android.media.MediaCodecList
import android.os.Build
import com.bureau.base.AppInfo
import com.bureau.devicefingerprint.datacollectors.AdvertisingProvider
import com.bureau.devicefingerprint.datacollectors.AndroidIdProvider
import com.bureau.devicefingerprint.datacollectors.AppSignaturesProvider
import com.bureau.devicefingerprint.datacollectors.BatteryInfoProvider
import com.bureau.devicefingerprint.datacollectors.CallInfoProvider
import com.bureau.devicefingerprint.datacollectors.CameraInfoProvider
import com.bureau.devicefingerprint.datacollectors.CameraInjectionInfoProvider
import com.bureau.devicefingerprint.datacollectors.CodecInfoProviderImpl
import com.bureau.devicefingerprint.datacollectors.CpuInfoProvider
import com.bureau.devicefingerprint.datacollectors.CpuInfoProviderImpl
import com.bureau.devicefingerprint.datacollectors.DebuggableInfoProvider
import com.bureau.devicefingerprint.datacollectors.DevicePersonalizationInfoProvider
import com.bureau.devicefingerprint.datacollectors.DeviceSecurityInfoProvider
import com.bureau.devicefingerprint.datacollectors.DeviceStorageProvider
import com.bureau.devicefingerprint.datacollectors.EmulatorPropertiesProvider
import com.bureau.devicefingerprint.datacollectors.FileProvisionsProvider
import com.bureau.devicefingerprint.datacollectors.FingerprintSensorInfoProvider
import com.bureau.devicefingerprint.datacollectors.GpuInfoProvider
import com.bureau.devicefingerprint.datacollectors.GsfIdProvider
import com.bureau.devicefingerprint.datacollectors.HookDetectionCheckProvider
import com.bureau.devicefingerprint.datacollectors.InputDeviceDataSource
import com.bureau.devicefingerprint.datacollectors.InstallInfoProvider
import com.bureau.devicefingerprint.datacollectors.KeyStoreProvider
import com.bureau.devicefingerprint.datacollectors.LocationInfoProvider
import com.bureau.devicefingerprint.datacollectors.MediaDrmIdProvider
import com.bureau.devicefingerprint.datacollectors.MemInfoProvider
import com.bureau.devicefingerprint.datacollectors.NetworkInfoProvider
import com.bureau.devicefingerprint.datacollectors.NewRawParametersProvider
import com.bureau.devicefingerprint.datacollectors.OsBuildInfoProvider
import com.bureau.devicefingerprint.datacollectors.PackageManagerDataSource
import com.bureau.devicefingerprint.datacollectors.ScreenInfoProvider
import com.bureau.devicefingerprint.datacollectors.SensorDataSource
import com.bureau.devicefingerprint.datacollectors.SettingsDataSource
import com.bureau.devicefingerprint.datacollectors.SimInfoProvider
import com.bureau.devicefingerprint.datacollectors.SocialEngineeringInfoProvider
import com.bureau.devicefingerprint.datacollectors.SystemPropertiesProvider
import com.bureau.devicefingerprint.datacollectors.TelecomBaseStationInfoProvider
import com.bureau.devicefingerprint.datacollectors.WebInfoProvider
import com.bureau.devicefingerprint.datacollectors.WifiInfoProvider
import com.bureau.devicefingerprint.di.DeviceFingerprintScope
import com.bureau.devicefingerprint.emulatordetection.EmulatorDetection
import com.bureau.devicefingerprint.models.CredentialBuilder
import com.bureau.devicefingerprint.models.devicedataholder.AndroidId
import com.bureau.devicefingerprint.models.devicedataholder.BatteryInfo
import com.bureau.devicefingerprint.models.devicedataholder.CameraCoreInfo
import com.bureau.devicefingerprint.models.devicedataholder.CameraInfo
import com.bureau.devicefingerprint.models.devicedataholder.CodecInfo
import com.bureau.devicefingerprint.models.devicedataholder.CpuInfo
import com.bureau.devicefingerprint.models.devicedataholder.CurrentLocation
import com.bureau.devicefingerprint.models.devicedataholder.DebuggerInfo
import com.bureau.devicefingerprint.models.devicedataholder.DeviceDataHolder
import com.bureau.devicefingerprint.models.devicedataholder.DevicePersonalizationInfo
import com.bureau.devicefingerprint.models.devicedataholder.DeviceSecurityInfo
import com.bureau.devicefingerprint.models.devicedataholder.EmulatorInfo
import com.bureau.devicefingerprint.models.devicedataholder.FingerPrintSensorInfo
import com.bureau.devicefingerprint.models.devicedataholder.GpuInfo
import com.bureau.devicefingerprint.models.devicedataholder.GsfId
import com.bureau.devicefingerprint.models.devicedataholder.InputDeviceData
import com.bureau.devicefingerprint.models.devicedataholder.InputDevicesInfo
import com.bureau.devicefingerprint.models.devicedataholder.LastKnownLocation
import com.bureau.devicefingerprint.models.devicedataholder.LocationInfo
import com.bureau.devicefingerprint.models.devicedataholder.MediaDrmId
import com.bureau.devicefingerprint.models.devicedataholder.MemInfo
import com.bureau.devicefingerprint.models.devicedataholder.NdkInfo
import com.bureau.devicefingerprint.models.devicedataholder.NetworkInfo
import com.bureau.devicefingerprint.models.devicedataholder.OsBuildInfo
import com.bureau.devicefingerprint.models.devicedataholder.PackageManagerInfo
import com.bureau.devicefingerprint.models.devicedataholder.PastRealTimeDatHolder
import com.bureau.devicefingerprint.models.devicedataholder.RawParameters
import com.bureau.devicefingerprint.models.devicedataholder.ScreenInfoHolder
import com.bureau.devicefingerprint.models.devicedataholder.SensorData
import com.bureau.devicefingerprint.models.devicedataholder.SensorsInfo
import com.bureau.devicefingerprint.models.devicedataholder.SettingsInfo
import com.bureau.devicefingerprint.models.devicedataholder.SignaturesInfo
import com.bureau.devicefingerprint.models.devicedataholder.SimInfo
import com.bureau.devicefingerprint.models.devicedataholder.SocialEngineeringInfo
import com.bureau.devicefingerprint.models.devicedataholder.SystemApplication
import com.bureau.devicefingerprint.models.devicedataholder.SystemProperties
import com.bureau.devicefingerprint.models.devicedataholder.UserApplication
import com.bureau.devicefingerprint.models.devicedataholder.WebInfo
import com.bureau.devicefingerprint.models.devicedataholder.WifiInfoParameters
import com.bureau.devicefingerprint.tools.DependencyHelpers
import com.bureau.devicefingerprint.tools.NDKMappers
import com.bureau.devicefingerprint.tools.SharedPreferenceHelper
import com.bureau.devicefingerprint.tools.executeSafe
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.withContext
import java.io.UnsupportedEncodingException
import java.nio.charset.Charset
import javax.inject.Inject

@DeviceFingerprintScope
class DataMapper @Inject constructor(
    private val context: Context,
    private val androidIdProvider: AndroidIdProvider,
    private val batteryInfoProvider: BatteryInfoProvider,
    private val cameraInfoProvider: CameraInfoProvider,
    private val osBuildInfoProvider: OsBuildInfoProvider,
    private val deviceStorageProvider: DeviceStorageProvider,
    private val cpuInfoProvider: CpuInfoProvider,
    private val devicePersonalizationInfoProvider: DevicePersonalizationInfoProvider,
    private val deviceSecurityInfoProvider: DeviceSecurityInfoProvider,
    private val fingerprintSensorInfoProvider: FingerprintSensorInfoProvider,
    private val gpuInfoProvider: GpuInfoProvider,
    private val gsfIdProvider: GsfIdProvider,
    private val inputDeviceDataSource: InputDeviceDataSource,
    private val packageManagerDataSource: PackageManagerDataSource,
    private val sensorDataSource: SensorDataSource,
    private val settingsDataSource: SettingsDataSource,
    private val emulatorPropertiesProvider: EmulatorPropertiesProvider,
    private val systemPropertiesProvider: SystemPropertiesProvider,
    private val networkInfoProvider: NetworkInfoProvider,
    private val webInfoProvider: WebInfoProvider,
    private val keyStoreProvider: KeyStoreProvider,
    private val debuggableInfoProvider: DebuggableInfoProvider,
    private val appSignaturesProvider: AppSignaturesProvider,
    private val callInfoProvider: CallInfoProvider,
    private val screenInfoProvider: ScreenInfoProvider,
    private val telecomBaseStationInfoProvider: TelecomBaseStationInfoProvider,
    private val newRawParametersProvider: NewRawParametersProvider,
    private val socialEngineeringInfoProvider: SocialEngineeringInfoProvider,
    private val wifiInfoProvider: WifiInfoProvider,
    private val cameraInjectionInfoProvider: CameraInjectionInfoProvider,
    private val appInfo: AppInfo,
    private val locationInfoProvider: LocationInfoProvider,
    private val simInfoProvider: SimInfoProvider,
    private val hookDetectionCheckProvider: HookDetectionCheckProvider,
    private val fileProvisionsProvider: FileProvisionsProvider,
    private val advertisingProvider: AdvertisingProvider,
    private val installInfoProvider: InstallInfoProvider
) {

    private var ndkMappers: NDKMappers = NDKMappers()
    private var hasLocationPermission = false
    private var vendorName = ""

    fun getAsyncOSData(
        credentialBuilder: CredentialBuilder?,
        deviceDataHolder: DeviceDataHolder
    ): DeviceDataHolder {
        getNDKData(deviceDataHolder)

        deviceDataHolder.androidId_ =
            AndroidId(androidIdProvider.getAndroidId())

        deviceDataHolder.batteryInfo_ = BatteryInfo(
            batteryCapacity_ = batteryInfoProvider.batteryTotalCapacity(),
            batteryHealth_ = batteryInfoProvider.batteryHealth(),
            isCharging_ = batteryInfoProvider.isCharging()
        )

        deviceDataHolder.cameraInfo_ = getCameraInfoData()
        deviceDataHolder.elapsedTime_ = osBuildInfoProvider.getElapsedTime()
        deviceDataHolder.encryptionStatus_ = deviceStorageProvider.getEncryptionStatus()
        deviceDataHolder.cpuInfo_ = getCpuInfoData()
        deviceDataHolder.devicePersonalizationInfo_ = getDevicePersonalizationInfo()
        deviceDataHolder.deviceSecurityInfo_ = getDeviceSecurityInfo()
        deviceDataHolder.fingerPrintSensorInfo_ = FingerPrintSensorInfo(
            sensorStatus_ = fingerprintSensorInfoProvider.getStatus().name
        )
        deviceDataHolder.gpuInfo_ =
            GpuInfo(glesVersion_ = gpuInfoProvider.glesVersion())
        deviceDataHolder.gsfId_ =
            GsfId(gsfId_ = gsfIdProvider.getGsfAndroidId())
        deviceDataHolder.inputDevicesInfo_ = getInputDevicesInfo()

        val memInfoProvider = DependencyHelpers.createMemoryInfoProvider(context)
        deviceDataHolder.memInfo_ = getMemInfoData(memInfoProvider)

        deviceDataHolder.osBuildInfo_ = getOSBuildData(osBuildInfoProvider)
        deviceDataHolder.packageManagerInfo_ = getPackageManagerInfoData()
        deviceDataHolder.sensorsInfo_ = getSensorInfoData()
        deviceDataHolder.settingsInfo_ = getSettingsInfoData()
        deviceDataHolder.emulatorInfo_ = EmulatorInfo(
            emulatorIp_ = emulatorPropertiesProvider.getEmulatorIP(),
            qEMUVendor_ = vendorName
        )
        deviceDataHolder.systemProperties_ = getSystemProperties()
        deviceDataHolder.networkInfo_ = NetworkInfo(
            iPV4_ = networkInfoProvider.getIpv4Address(),
            iPV6_ = networkInfoProvider.getIpv6Address(),
            ethernetMACAddress_ = networkInfoProvider.getEthernetMACAddress(),
            wifiMacAddress_ = networkInfoProvider.getWifiMACAddress(),
            isVpnConnected_ = networkInfoProvider.isConnectedToVPN(),
            isWifiConnected_ = networkInfoProvider.getIsWifiConnected(),
            isCellularConnected_ = networkInfoProvider.getIsCellularConnected(),
            networkType = networkInfoProvider.getNetworkType()
        )
        deviceDataHolder.webInfo_ = WebInfo(userAgent_ = webInfoProvider.getUserAgent())
        deviceDataHolder.keyStoreKey_ = keyStoreProvider.getKey()
        deviceDataHolder.userId_ = credentialBuilder?.userId ?: ""
        deviceDataHolder.flow_ = credentialBuilder?.flow ?: ""
        deviceDataHolder.sdkVersion_ = BuildConfig.LIBRARY_VERSION_NAME
        deviceDataHolder.debuggerInfo_ = DebuggerInfo(
            isDebuggerAttached_ = debuggableInfoProvider.isDebuggerAttached(),
            isRooted_ = debuggableInfoProvider.isRooted()
        )
        deviceDataHolder.signaturesInfo_ = getSignaturesInfo()
        deviceDataHolder.sessionId_ = credentialBuilder?.sessionId ?: ""
        deviceDataHolder.voiceCallMode_ = callInfoProvider.audioMode()
        deviceDataHolder.appDirectorypath_ = context.filesDir.absolutePath
        deviceDataHolder.screenInfo_ = ScreenInfoHolder(
            screenCount_ = screenInfoProvider.getScreenCount(),
            drawableDensity_ = screenInfoProvider.getDrawableDensity(),
            layoutSize_ = screenInfoProvider.getLayoutSize(),
            layoutHeight_ = screenInfoProvider.getLayoutHeight(),
            layoutWidth_ = screenInfoProvider.getLayoutWidth()
        )
        deviceDataHolder.telecomBaseStationInfo_ = telecomBaseStationInfoProvider.getTelecomBaseStationInfo()
        deviceDataHolder.rawParameters = RawParameters(
            isMuted = newRawParametersProvider.isMuted(),
            volumeLevel = newRawParametersProvider.getVolumePercentage(),
            batteryLevel = batteryInfoProvider.getBatteryLevel(),
            batteryTemperatureInCentigrade = batteryInfoProvider.getBatterTemperature(),
            batteryVoltage = batteryInfoProvider.getBatteryVoltage(),
            buildId = osBuildInfoProvider.getBuildId(),
            buildTime = osBuildInfoProvider.getBuildTime(),
            freeExternalStorageSpace = memInfoProvider.getFreeExternalStorageSpace(),
            freeInternalStorageSpace = memInfoProvider.getFreeInternalStorageSpace(),
            brightness = newRawParametersProvider.getBrightness()
        )

        deviceDataHolder.socialEngineeringInfo = SocialEngineeringInfo(isSocialEngineeringAppInstalled = socialEngineeringInfoProvider.getIsSocialEngineeringAppInstalled())

        deviceDataHolder.wifiInfoParameters = WifiInfoParameters(wifiInfoProvider.getSSID(), wifiInfoProvider.getBSSID(), wifiInfoProvider.getRSSI())

        deviceDataHolder.isEmulatorDetectedAdvanced = EmulatorDetection.isEmulatorDetected(context)

        deviceDataHolder.cameraInjectionInfo = cameraInjectionInfoProvider.getCameraInjection()

        deviceDataHolder.metaData.clientInfo.run {
            clientAppPackage = appInfo.getPackageName()
            clientAppName = appInfo.getAppName()
            clientAppVersion = appInfo.getVersion()
        }

        deviceDataHolder.simInfo = SimInfo(numOfSimSlots = simInfoProvider.getNumOfSimSlots(), numOfSims = simInfoProvider.getNumOfSimsPresent(), isDataRoamingEnabled = simInfoProvider.isDataRoamingEnabled())

        return deviceDataHolder
    }

    fun getMediaDrmId(): MediaDrmId {
        return MediaDrmId(drmId_ = MediaDrmIdProvider().getMediaDrmId())
    }

    fun getMediaCodecInfo(): CodecInfo {
        return getCodecInfoData(MediaCodecList(MediaCodecList.ALL_CODECS))
    }

    fun getIsHookingDetected(): Boolean {
        return hookDetectionCheckProvider.isHookingDetected()
    }

    fun addGeocodeData(deviceData: DeviceDataHolder): DeviceDataHolder {
        if (locationInfoProvider.decodeCityName(locationInfoProvider.currentLocation).isEmpty()) {
            deviceData.locationInfo_?.currentLocation_?.city_ = locationInfoProvider.decodeCityName(
                locationInfoProvider.lastKnownLocation
            )
        } else {
            deviceData.locationInfo_?.currentLocation_?.city_ = locationInfoProvider.decodeCityName(
                locationInfoProvider.currentLocation
            )
        }

        deviceData.locationInfo_?.lastKnownLocation_?.city_ =
            locationInfoProvider.decodeCityName(locationInfoProvider.lastKnownLocation)
        if (locationInfoProvider.decodeCountryName(locationInfoProvider.currentLocation).isEmpty()) {
            deviceData.locationInfo_?.currentLocation_?.country_ = locationInfoProvider.decodeCountryName(
                locationInfoProvider.lastKnownLocation
            )
        } else {
            deviceData.locationInfo_?.currentLocation_?.country_ = locationInfoProvider.decodeCountryName(
                locationInfoProvider.currentLocation
            )
        }

        deviceData.locationInfo_?.lastKnownLocation_?.country_ =
            locationInfoProvider.decodeCountryName(locationInfoProvider.lastKnownLocation)
        if (locationInfoProvider.decodeRegionName(locationInfoProvider.currentLocation).isEmpty()) {
            deviceData.locationInfo_?.currentLocation_?.region_ = locationInfoProvider.decodeRegionName(
                locationInfoProvider.lastKnownLocation
            )
        } else {
            deviceData.locationInfo_?.currentLocation_?.region_ =
                locationInfoProvider.decodeRegionName(locationInfoProvider.currentLocation)
        }

        deviceData.locationInfo_?.lastKnownLocation_?.region_ =
            locationInfoProvider.decodeRegionName(locationInfoProvider.lastKnownLocation)
        return deviceData
    }

    fun getSynchronisedOSData(
        hasLocationPermission: Boolean,
        deviceData: DeviceDataHolder
    ) {
        this.hasLocationPermission = hasLocationPermission
        deviceData.locationInfo_ = getLocationInfo()

        val (gpsEnabled, networkEnabled) = locationInfoProvider.getAreGpsAndNetworkEnabled()
        deviceData.metaData.gpsEnabled = gpsEnabled
        deviceData.metaData.networkEnabled = networkEnabled

        deviceData.uniqueFileId_ = fileProvisionsProvider.getUniqueIDFromFile()
    }

    suspend fun getAsynchronousIds(deviceData: DeviceDataHolder): DeviceDataHolder = withContext(Dispatchers.IO) {
        deviceData.advertisingID_ = advertisingProvider.getAdvertisingId()
        deviceData
    }

    private fun getSignaturesInfo(): SignaturesInfo {
        val list: ArrayList<String> = ArrayList()
        val signatures: Array<Signature>? = appSignaturesProvider.getSignatures()

        signatures?.forEach {
            list.add(it.toCharsString())
        }
        return SignaturesInfo(
            isPlayStoreInstall_ = installInfoProvider.getIsPlayStoreInstall(),
            signatureHashcode_ = list
        )
    }

    private fun getNDKData(deviceDataHolder: DeviceDataHolder) {
        deviceDataHolder.ndkInfo_ =
            NdkInfo(
                applicationId_ = executeSafe({ ndkMappers.getApplicationId() }, ""),
                installPath_ = executeSafe({ ndkMappers.getSourceDir(context) }, "")
            )
        val byteArray: ByteArray = ndkMappers.getEmulatorVendor()
        try {
            val nativeString = String(byteArray, Charset.defaultCharset())
            vendorName = nativeString
        } catch (e: UnsupportedEncodingException) {
            BureauAPI.handleException(e)
            e.printStackTrace()
        }
    }

    private fun getCameraInfoData(): CameraInfo {
        val list: ArrayList<CameraCoreInfo> = ArrayList()
        cameraInfoProvider.getCameraInfo().forEach {
            list.add(
                CameraCoreInfo(
                    cameraName_ = it.cameraName,
                    cameraOrientation_ = it.cameraOrientation,
                    cameraType_ = it.cameraType
                )
            )
        }
        return CameraInfo(cameraCoreInfo_ = list)
    }

    private fun getCodecInfoData(codecList: MediaCodecList): CodecInfo {
        val codecInfo = CodecInfo()

        CodecInfoProviderImpl(codecList).codecsList().forEach {
            codecInfo.name_ = it.name
            val list: ArrayList<String> = ArrayList()
            it.capabilities.forEach { it2 ->
                list.add(it2)
            }
            codecInfo.capabilities_ = list
        }
        return codecInfo
    }

    private fun getCpuInfoData(): CpuInfo {
        val map: HashMap<String, String> = HashMap()
        cpuInfoProvider.cpuInfo().asIterable().forEach {
            map[it.key] = it.value
        }

        return CpuInfo(
            abiType_ = CpuInfoProviderImpl().abiType(),
            coreCount_ = CpuInfoProviderImpl().coresCount(),
            map
        )
    }

    private fun getDevicePersonalizationInfo(): DevicePersonalizationInfo {
        val list: ArrayList<String> = ArrayList()

        devicePersonalizationInfoProvider.availableLocales().forEach {
            list.add(it)
        }
        return DevicePersonalizationInfo(
            defaultLanguage_ = devicePersonalizationInfoProvider.defaultLanguage(),
            regionOrCountry_ = devicePersonalizationInfoProvider.regionCountry(),
            ringtoneSource_ = devicePersonalizationInfoProvider.ringtoneSource(),
            timeZone_ = devicePersonalizationInfoProvider.timezone(),
            availableLocales_ = list
        )
    }

    private fun getDeviceSecurityInfo(): DeviceSecurityInfo {
        val map: HashMap<String, String> = HashMap()

        deviceSecurityInfoProvider.securityProvidersToMap().asIterable().forEach {
            map[it.key] = it.value
        }

        return DeviceSecurityInfo(
            encryptionStatus_ = deviceSecurityInfoProvider.encryptionStatus(),
            isPinEnabled_ = deviceSecurityInfoProvider.isPinSecurityEnabled(),
            securityProviders_ = map
        )
    }

    private fun getInputDevicesInfo(): InputDevicesInfo {
        val list: ArrayList<InputDeviceData> = ArrayList()

        inputDeviceDataSource.getInputDeviceData().forEach {
            list.add(InputDeviceData(name_ = it.name, vendor_ = it.vendor))
        }
        return InputDevicesInfo(inputDeviceData_ = list)
    }

    private fun getMemInfoData(memInfoProvider: MemInfoProvider): MemInfo {
        return MemInfo(
            totalRAM_ = memInfoProvider.totalRAM(),
            totalExternalStorageSpace_ = memInfoProvider.totalExternalStorageSpace(),
            totalInternalStorageSpace_ = memInfoProvider.totalInternalStorageSpace()
        )
    }

    private fun getOSBuildData(provider: OsBuildInfoProvider): OsBuildInfo {
        return OsBuildInfo(
            androidVersion_ = provider.androidVersion(),
            fingerprint_ = provider.fingerprint(),
            kernelVersion_ = provider.kernelVersion(),
            manufacturerName_ = provider.manufacturerName(),
            modelName_ = provider.modelName(),
            sdkVersion_ = provider.sdkVersion(),
            device_ = provider.device(),
            product_ = provider.product(),
            host_ = provider.host(),
            brand_ = provider.brand(),
            board_ = provider.board()
        )
    }

    private fun getPackageManagerInfoData(): PackageManagerInfo {
        val userList: ArrayList<UserApplication> = ArrayList()
        val systemList: ArrayList<SystemApplication> = ArrayList()

        if (Build.VERSION.SDK_INT <= Build.VERSION_CODES.Q || BureauAPI.hasPermissions(
                context,
                listOf(Manifest.permission.QUERY_ALL_PACKAGES)
            )
        ) {
            packageManagerDataSource.getSystemApplicationsList().forEach {
                systemList.add(SystemApplication(it.packageName, it.installTime))
            }
            packageManagerDataSource.getApplicationsList().forEach {
                userList.add(UserApplication(it.packageName, it.installTime))
            }
        }

        return PackageManagerInfo(
            packageVersion_ = packageManagerDataSource.getPackageVersion(),
            isPackageDebuggable_ = debuggableInfoProvider.isDebuggable(),
            systemApplicationList_ = systemList,
            userApplicationList_ = userList,
            clientAppInstallTime = packageManagerDataSource.getClientAppInstallTime()
        )
    }

    private fun getSensorInfoData(): SensorsInfo {
        val list: ArrayList<SensorData> = ArrayList()

        sensorDataSource.sensors().forEach {
            list.add(SensorData(sensorName_ = it.sensorName, sensorVendorName_ = it.vendorName))
        }

        return SensorsInfo(sensorData_ = list)
    }

    private fun getSettingsInfoData(): SettingsInfo {
        return SettingsInfo(
            accessibilityEnabled_ = settingsDataSource.accessibilityEnabled(),
            adbEnabled_ = settingsDataSource.adbEnabled(),
            alarmAlertPath_ = settingsDataSource.alarmAlertPath(),
            dataRoamingEnabled_ = settingsDataSource.dataRoamingEnabled(),
            dateFormat_ = settingsDataSource.dateFormat(),
            defaultInputMethod_ = settingsDataSource.defaultInputMethod(),
            developmentSettingsEnabled_ = settingsDataSource.developmentSettingsEnabled(),
            endButtonBehaviour_ = settingsDataSource.endButtonBehaviour(),
            fontScale_ = settingsDataSource.fontScale(),
            httpProxy_ = settingsDataSource.httpProxy(),
            rttCallingMode_ = settingsDataSource.rttCallingMode(),
            screenOffTimeout_ = settingsDataSource.screenOffTimeout(),
            textAutoPunctuate_ = settingsDataSource.textAutoPunctuate(),
            textAutoReplaceEnable_ = settingsDataSource.textAutoReplaceEnable(),
            touchExplorationEnabled_ = settingsDataSource.touchExplorationEnabled(),
            transitionAnimationScale_ = settingsDataSource.transitionAnimationScale(),
            windowAnimationScale_ = settingsDataSource.windowAnimationScale()
        )
    }

    private fun getSystemProperties(): SystemProperties {
        return SystemProperties(
            attributeData_ = systemPropertiesProvider.getAttributeData(),
            bootMode_ = systemPropertiesProvider.getBootMode(),
            density_ = systemPropertiesProvider.getLCDDensity(),
            deviceProps_ = systemPropertiesProvider.getProps(),
            fakeCamera_ = systemPropertiesProvider.getFakeCamera(),
            kernelGlESEnabled_ = systemPropertiesProvider.getGLES(),
            kernelVMRunning_ = systemPropertiesProvider.getKernelData(),
            mainKeyEnabled_ = systemPropertiesProvider.getMainKey(),
            serialNo_ = systemPropertiesProvider.getSerialNo(),
            kernelQEMUEnabled_ = systemPropertiesProvider.getQEMU(),
            systemProperties = systemPropertiesProvider.getSystemProperties()
        )
    }

    private fun getLocationInfo(): LocationInfo {
        locationInfoProvider.getKnownLocation()

        return LocationInfo(
            currentLocation_ = CurrentLocation(
                latitude_ = LocationInfoProvider.locationLatitude.ifEmpty { locationInfoProvider.getLastKnownLocation().latitude },
                longitude_ = LocationInfoProvider.locationLongitude.ifEmpty { locationInfoProvider.getLastKnownLocation().longitude },
                gpsPrecision_ = if (LocationInfoProvider.precision.equals(0f)) locationInfoProvider.getLastKnownLocation().locationPrecision else LocationInfoProvider.precision
            ),
            lastKnownLocation_ = LastKnownLocation(
                latitude_ = locationInfoProvider.getLastKnownLocation().latitude,
                longitude_ = locationInfoProvider.getLastKnownLocation().longitude,
                gpsPrecision_ = if (LocationInfoProvider.precision.equals(0f)) locationInfoProvider.getLastKnownLocation().locationPrecision else LocationInfoProvider.precision
            ),
            isLocationEnabled_ = locationInfoProvider.getIsLocationEnabled(),
            isLocationPermissionEnabled_ = hasLocationPermission,
            isMockGPS_ = locationInfoProvider.getLastKnownLocation().isMockLocation
        ).also {
            locationInfoProvider.removeListener()
        }
    }

    fun addPastRealData(deviceDataHolder: DeviceDataHolder): DeviceDataHolder {
        deviceDataHolder.pastRealTimeDataInfo_ = PastRealTimeDatHolder(
            pastVpnStatus_ = SharedPreferenceHelper.getSharedPreferenceBoolean(
                context,
                SharedPreferenceHelper.isPastVPNData,
                defValue = false,
                isDataMonitoring = true
            ),
            pastMockGPSStatus_ = SharedPreferenceHelper.getSharedPreferenceBoolean(
                context,
                SharedPreferenceHelper.isPastMockGPSData,
                defValue = false,
                isDataMonitoring = true
            ),
            pastDebuggerStatus_ = SharedPreferenceHelper.getSharedPreferenceBoolean(
                context,
                SharedPreferenceHelper.isPastDebuggerData,
                defValue = false,
                isDataMonitoring = true
            )
        )
        return deviceDataHolder
    }
}
